#!/usr/bin/python

import gzip
import tarfile
import zipfile

import tempfile
import glob
import os
import os.path
import shutil
import argparse
import re
from enum import Enum

Lms = Enum('LMS', ['Canvas', 'ZyBooks'])
CANVAS_PATTERN = r"(?P<name>[^_]*)(?P<late>_LATE)?_(?P<lms_id>[0-9]*)_(?P<sub_id>[0-9]*)_(?P<filename>.*)"


def main():
    parser = argparse.ArgumentParser(description="Unzips submissions from LMS. Defaults to Canvas format.", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_help = True
    parser.add_argument('filename', help='submissions zip file')
    parser.add_argument('destination', help='where to extract submissions to')
    parser.add_argument('-z', '--zybooks', help='process ZyBooks archive', action='store_true')

    args = parser.parse_args()
    filetype = (Lms.ZyBooks if args.zybooks else Lms.Canvas) # Default to Canvas file type

    # Set up a temporary working directory for submission data and extract the archive.
    tmp_path = tempfile.mkdtemp()
    with zipfile.ZipFile(args.filename) as zf:
        zf.extractall(path=tmp_path)

    # Go through the raw submission files one at a time.
    submissions = glob.glob(os.path.join(tmp_path, "*"))
    success_count = 0
    warning_count = 0

    for rawfile in submissions:
        success_count += 1

        # separate the submission name from the file type (for later use)
        submission, ext = os.path.splitext(rawfile)

        # Handle the special case of gzip files (which by default contain another extension)
        while ext and ext.lower().split('.')[1] == "gz":
           submission, ext2 = os.path.splitext(submission)
           ext = ext2 + ext

        # Based in the LMS type, extract submission information from the submission string.
        if filetype == Lms.Canvas:
            attributes = re.match(CANVAS_PATTERN, os.path.basename(submission))

            # Deal with Canvas appending "LATE" to student names with an underscore (add to name with a dash)
            student = attributes.group('name')
            lms_id = attributes.group('lms_id')
            submission_id = attributes.group('sub_id')
            header = attributes.group('filename')
            late = (True if attributes.group('late') else False)

            # Give a warning if there was trouble parsing the name.
            if not (student or lms_id or submission_id or header):
                print("Warning: missing elements in [%s]. Rebuilt files may be malformed." % submission)

            # Account for Canvas's submission count suffix (gross, Canvas).
            count_info = header.rsplit("-", 1)
            if len(count_info) > 1:
                try:
                    int(count_info[1])
                    header = count_info[0]
                except ValueError:
                    pass # Right side of "-" was not a number, so this is not a Canvas submission count suffix.

        # Handler for ZyBooks submissions - TODO: Redo with regex (as above)
        elif filetype == Lms.ZyBooks:
            attributes = os.path.basename(submission).rsplit("_", 3)

            # Handle missing elements in filenames as best we can (limited to weird Canvas output configurations)
            if len(attributes) < 4:
                print("Warning: [%s] cannot be reformatted. Assuming generic values." % submission)
                attributes[1:1] = [""] * 4 - len(attributes)

            student, lms_id, submission_date, submission_time = attributes
            submission_id = submission_date + "." + submission_time
            header = "submission"

        # We shouldn't reach this. If we do, something is wrong; it should catch invalid LMS types.
        else:
            print("Uh-oh, invalid LMS type... we should never get here!")
            exit(1)

        # If the submission was late, mark it as such.
        if attributes.group('late'):
            student += "-LATE"

        # Now that we've processed submission information, reconstruct the files.
        ext = ext.lower()
        filename = header + ext

        student_path = os.path.join(args.destination, student + "_" + lms_id)

        if not os.path.isdir(student_path):
            os.makedirs(student_path)

        if ext == ".zip":
            try:
                with zipfile.ZipFile(rawfile) as zf:
                    zf.extractall(path=student_path)
            except zipfile.BadZipFile:
                print("Warning: [%s] is not a zipfile; treating like normal file." % os.path.basename(rawfile))
                shutil.copyfile(rawfile, os.path.join(student_path, filename))
                warning_count += 1
        elif ext == ".tgz" or ext == ".tar.gz" or ext.endswith(".tar"):
            try:
                with tarfile.open(rawfile, ("r" if ext.endswith(".tar") else "r:gz")) as tf:
                    tf.extractall(path=student_path)
            except tarfile.TarError:
                print("Warning: [%s] could not be read / extracted; treating like normal file." % os.path.basename(rawfile))
                shutil.copyfile(rawfile, os.path.join(student_path, filename))
                warning_count += 1
        elif ext.lower().endswith(".gz"):
            try:
                with gzip.open(rawfile, "rb") as gzf:
                    with open(os.path.join(student_path, filename.rsplit(".", 1)[0]), "wb") as outfile:
                        outfile.write(gzf.read())
            except gzip.BadGzipFile:
                print("Warning: [%s] could not be read / extracted; treating like normal file." % os.path.basename(rawfile))
                shutil.copyfile(rawfile, os.path.join(student_path, filename))
                warning_count += 1
            except:
                print("Warning: could not write extracted file; copying [%s] as normal file." % os.path.basename(rawfile))
                shutil.copyfile(rawfile, os.path.join(student_path, filename))
                warning_count += 1
        else:
            shutil.copyfile(rawfile, os.path.join(student_path, filename))

    print("Successfuly extracted %d submissions with %d warnings." % (success_count, warning_count))

if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Copyright (c) 2017 Cacti Council Inc., 2018-2020 University of Florida

import argparse
import glob
import os
import shutil
import sys
import subprocess
import time
import os.path
import traceback
import numbers
import string
import logging
import dill
import pathos.pools as pools

from . import toolbox
from . import VERSION
from concurrent import futures


__NULL_PATH__ = "_!$NULL_PATH$!_"

# handle command line args
def parse_arguments():
    parser = argparse.ArgumentParser(description='A program to run a set of tests for a programming assignment.')
    parser.add_help = True
    parser.add_argument('suite_path', nargs='?', default="./", help='path of test suite to load')
    parser.add_argument('target_path', nargs='?', default="Projects", help='path of target projects (by subdirectory)')
    parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + VERSION)
    parser.add_argument('-t', '--threaded', dest='threaded', action='store_true', help='use threads instead of processes')
    parser.add_argument('-q', '--quiet', dest='INFO', action='store_false', help='execute in quiet mode (console)')
    parser.add_argument('-w', '--warn', dest='WARN', action='store_true', help='display warning information (console)')
    parser.add_argument('-d', '--debug', dest='DEBUG', action='store_true', help='capture debug information (logfile)')
    parser.add_argument('-s', '--set', dest='set', default="*", help = 'test only projects designated (e.g., *_LATE*)')
    parser.add_argument('-T', '--tests', nargs=2, dest='set_tests', action="append", metavar=('test_set','test_list'),
                        default=[], help='testset & tests to run, e.g.: "MySet 1,2,0" (comma-separated); default: all')

    config = parser.parse_args(sys.argv[1:])
    set_test_mapping = {}

    if config.set_tests:
        for set_id, num_list in config.set_tests:
            if set_id in set_test_mapping:
                print("WARNING: repeated set_id '%s' in test list parameter. Skipping." % set_id)
                continue

            try:
                num_list = [int(testnum.strip()) for testnum in num_list.split(',')]
                set_test_mapping[set_id] = num_list
            except:
                print("WARNING: Skipping set test list [%s: %s] - incorrect syntax;" % (set_id, num_list))
                print("Must be comma separated integer list with no spaces (e.g., '4,5,0')")
                continue

    config.logformat = "%(message)s"
    config.set_tests = set_test_mapping
    return config


def build_project(source_root, build_root, build_cfg, is_framework=False):
    result_error = None
    error_output = None
    current_dir = os.getcwd()

    # If there is a not a specified build directory, fall back to the source directory instead (if possible).
    if not build_root:
        build_root = source_root
        if source_root:
            logging.info("assuming combo build-and-source path...")
        else:
            logging.info("assuming directory-independent build steps...")

    # If the build exists, remove old project folder, recreate it, and switch to it.
    if build_root:
        if os.path.isdir(build_root) and not build_root == source_root:
            shutil.rmtree(build_root)
        if not os.path.exists(build_root):
            os.makedirs(build_root)
        os.chdir(build_root)

    # Prepare to make substitutions to the prep / build commands if applicable. Remove empty elements.
    command_sets = ('prep_cmd', 'compile_cmd', 'post_cmd')
    subs = {key : value for key, value in build_cfg.__dict__.items() if not key in command_sets}
    subs["source_dir"] = (source_root if source_root else __NULL_PATH__)
    subs["build_dir"] = (build_root if build_root else __NULL_PATH__)

    try:
        for entry in command_sets:
            if is_framework and entry in build_cfg.framework_flags and not build_cfg.framework_flags[entry]:
                continue # We are building the framework, and this command set flag is off (False). Skip it.

            if entry in build_cfg and build_cfg[entry]: # If the entry exists and is non-null, replace it.
                cmds = toolbox.replace(subs, build_cfg[entry], is_recursive=True, delete_empties=True)

                for cmd in cmds:
                    if any([__NULL_PATH__ in entry for entry in cmd]):
                        logging.info("\nSkipping entry without path:\n%s\n" % cmd)
                    else:
                        subprocess.check_output(cmd, stderr=subprocess.STDOUT, text=True)

    except subprocess.CalledProcessError as error:
        result_error = error
        error_output = error.output

    except FileNotFoundError as error:
        result_error = error

    os.chdir(current_dir)
    return result_error, error_output


def run_suite_tests(subject, framework, cfg):
    results = []
    exception_sets = {}

    # Run each project's tests.
    for test_set in cfg.sets:
        data_set, score, penalty_totals, exception_list = run_test_set(test_set, subject, framework, cfg)
        exception_sets[test_set.name] = exception_list

        # If the project didn't compile, just add a single line indicating that.
        if isinstance(data_set, str):
            exception_sets[test_set.name].append(data_set)
            data_set = [[], ["Grade: 0 (Does not compile / run)"]]
            score = 0

        else:
            # Add info on the score and penalty values for the project.
            data_set += [ [], ["Test Cases: %.2f (%.2f%%)" % (score * test_set.max_score, score * 100) ] ]
            overall_penalty = 0

            for penalty_num, case_penalty in enumerate(test_set.case_penalties):
                penalty_name, magnitude, _ = case_penalty
                overall_penalty += penalty_totals[penalty_num]
                data_set += [ [ "%s Penalty (overall): %.2f%%" % (penalty_name, penalty_totals[penalty_num] * 100) ] ]

            for penalty_num, set_penalty in enumerate(test_set.set_penalties):
                penalty_name, magnitude, _ = test_set.set_penalties[penalty_num]
                penalty_ind = penalty_num + len(test_set.case_penalties)
                overall_penalty += penalty_totals[penalty_ind]
                data_set += [ [ "%s Penalty (overall): %.2f%%" % (penalty_name, penalty_totals[penalty_ind] * 100) ] ]

            # Apply the penalties and scale to the number of points
            score = (score - min(test_set.max_penalty, overall_penalty)) * test_set.max_score

        # Add to the results list.
        data_set = [["Test-Set %s" % test_set.name]] + data_set
        results.append((test_set.name, score, data_set))

    return results, exception_sets


def run_test_set(test_set, subject, framework, cfg):
    # Prepare data structures and initialize the test set.
    exception_list = []
    set_context = cfg.initialize_test_set(test_set, subject, framework)
    num_of_total_tests = test_set.get_num_tests(set_context, subject, framework, cfg)
    penalty_totals = [0] * (len(test_set.case_penalties) + len(test_set.set_penalties))
    score = 0

    # Get the total number of tests. If none exist, return an error message.
    if not isinstance(num_of_total_tests, int) or num_of_total_tests == 0:
        cfg.shutdown_test_set(set_context)
        return "get_number_of_tests() returned [%s]" % num_of_total_tests, None, None, exception_list

    # If there are no tests identified yet to be run, generate a default list of all tests.
    tests_to_run = cfg.runtime.set_tests.get(test_set.id, list(range(0, num_of_total_tests)))

    # Add the initial notes at the top (might adjust this later for 'pure' CSV output)
    data_set = [["Total number of tests in set: %d." % num_of_total_tests], ["For this run: %s" % tests_to_run], []]

    # Prepare the header.
    header = [ 'Test No.', 'Score', 'Message', 'Desc.' ]
    header.extend(["%s-Pen" % penalty[0] for penalty in test_set.case_penalties])
    data_set.append(header)

    # Run each test and scale the score accordingly.
    for test_num in tests_to_run:
        # Sanity check: is this test number actually among those in the test set? If not, skip it.
        if test_num >= num_of_total_tests:
            logging.info("Warning: %d is greater than total number of tests (%d). Skipping." % (test_num, num_of_total_tests))
            continue

        # Set up the row for this test and run it.
        row = [ '%d' % test_num ]
        try:
            case_result = test_set.run_case_test(test_num, set_context, subject, framework, cfg)
        except Exception as e:
            stack_trace = traceback.format_exc()
            exception_list.append("Test %d, %s: %s\n%s" % (test_num, type(e).__name__, e, stack_trace))
            case_result = 0

        # If we successfuly completed the run, this should be a number; otherwise, a message.
        if isinstance(case_result, numbers.Number):
            case_score = case_result
            message = None
        else:
            case_score = 0
            message = str(case_result)

        score += case_score
        # Add score, run message, and description as applicable
        row.append('%.2f%%' % (case_score * 100))
        row.append(message if message else '')
        row.append(test_set.get_test_desc(test_num, set_context, subject, framework, cfg) if round(case_score, 10) < 1 else '')

        if case_score == 0:
            data_set.append(row)
            continue

        # Go through each penalty and run it (if valid).
        for penalty_num, case_penalty in enumerate(test_set.case_penalties):
            penalty_name, magnitude, pen_function = case_penalty
            penalty = pen_function(penalty_num, test_num, set_context, subject, framework, cfg)
            penalty_totals[penalty_num] += penalty * magnitude * case_score / len(tests_to_run)
            row.append('%.2f%%' % (penalty * 100))

        # Add this test data to the data set.
        data_set.append(row)

    for penalty_num, set_penalty in enumerate(test_set.set_penalties):
        penalty_name, magnitude, pen_function = set_penalty
        penalty = pen_function(penalty_num, set_context, subject, framework, cfg) * score * magnitude
        penalty_totals[penalty_num + len(test_set.case_penalties)] = penalty

    # Return the data set, score (proportion), and penalty totals
    cfg.shutdown_test_set(set_context)
    return data_set, score / len(tests_to_run), penalty_totals, exception_list


# Initialize environment and build & initialize framework (only need to do this once.)
def prepare_environment_and_framework(cfg):
    # Run initialization command line commands (pre-build) if applicable.
    logging.info("Initializing environment...")
    if cfg.build.init_cmd:
        for cmd in cfg.build.init_cmd:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT, text=True)
        logging.info("done.\n")
    else:
        logging.info("no tasks; skipping.\n")

    # Run build prep, build, and post-build commands.
    if cfg.build.prep_cmd or cfg.build.compile_cmd or cfg.build.post_cmd:
        logging.info("Prepping / building framework environment... ")
        result_error, error_output = build_project(cfg.build.framework_src, cfg.build.framework_bin, cfg.build, is_framework=True)
        if result_error:
            error_text = "%s: %s\n" % (type(result_error).__name__, result_error)
            logging.error(error_output if error_output else error_text)
            logging.info(error_text)
            return

    logging.info("Initializing framework... ")
    framework_data = cfg.initialize_framework()
    logging.info("done.\n")
    return framework_data


def shutdown_framework_and_environment(cfg, framework_context):
    # Shutdown framework.
    cfg.shutdown_framework(framework_context)

    # Shutdown environment if applicable.
    logging.info("Shutting down environment...")
    if cfg.build.shutdown_cmd:
        for cmd in cfg.build.shutdown_cmd:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT, text=True)
        logging.info("done.\n")
    else:
        logging.info("no tasks; done.\n")


# For each submission, copy the base files, then the submission, into the destination folder.
def prepare_and_test_submission(submission, framework_context, cfg):
    # Because this might be in a new process, we wil need to reset the console logger when prep the project.
    console_logger = toolbox.SelectiveStreamHandler(INFO=cfg.runtime.INFO, WARNING=cfg.runtime.WARN, CRITICAL=True)
    logging.basicConfig(format=cfg.runtime.logformat, level=logging.DEBUG, handlers=[console_logger])
    root_logger = logging.getLogger('')
    console_logger.terminator = ""

    if not os.path.isdir(submission):
        return None

    # Create a new subject folder with base files in it - then copy oer the submission to be tested.
    if os.path.isdir(cfg.build.destination):
        shutil.rmtree(cfg.build.destination)

    os.makedirs(cfg.build.destination)
    if cfg.build.base:
        shutil.copytree(cfg.build.base, cfg.build.destination, dirs_exist_ok=True)
    shutil.copytree(submission, cfg.build.destination, dirs_exist_ok=True)

    # Build the project.
    setup_exceptions = []
    logging.info("Prepping / building project(s) for " + submission + "... ")
    if cfg.build.prep_cmd or cfg.build.compile_cmd or cfg.build.post_cmd:
        error, output = build_project(cfg.build.subject_src, cfg.build.subject_bin, cfg.build)

        if error:
            logging.info("error building (see logs)... ")
            error_text = "While prepping/building, %s: %s\n" % (type(error).__name__, error)
            error_text += output if output else ""
            setup_exceptions.append(error_text)

    logging.info("Initializing... ")
    subject_context = None
    starting_dir = os.getcwd()

    try:
        subject_context = cfg.initialize_subject(submission, framework_context)
    except Exception as error:
        logging.info("error initializing (see logs)... ")
        setup_exceptions.append("Error initializing subject %s - %s: %s" % (submission, type(error).__name__, error))

    os.chdir(starting_dir)
    logging.info("done.\n")

    starting_dir = os.getcwd()
    results, exception_sets = run_suite_tests(subject_context, framework_context, cfg)
    cfg.shutdown_subject(subject_context)
    os.chdir(starting_dir)
    if len(setup_exceptions) > 0:
        exception_sets["Setup"] = setup_exceptions

    return results, exception_sets


def main():
    dill.settings['recurse']=True
    runtime = parse_arguments()

    # Save the current folder and move to the test suite location.
    starting_dir = os.getcwd()
    os.chdir(runtime.suite_path)

    # Load the config file for this project.
    config_path = os.path.join("Settings","config.py")

    if not os.path.isdir("Settings"):
        sys.stderr.write("Error: no settings folder. Exiting...\n")
        return

    elif not os.path.isfile(config_path) or not (config := toolbox.load_module(config_path)):
        sys.stderr.write("Error: could not load configuration file. Exiting...\n")
        return

    # Load and format the configuration.
    cfg = config.get_suite_config(runtime)
    cfg.normalize()

    # Prepare result paths.
    if not os.path.isdir(cfg.general.result_path):
        os.mkdir(cfg.general.result_path)
    summary_path = os.path.join(cfg.general.result_path, cfg.general.summary_file)

    # Set up logging.
    console_logger = toolbox.SelectiveStreamHandler(INFO=cfg.runtime.INFO, WARNING=cfg.runtime.WARN, CRITICAL=True)
    logging.basicConfig(format=cfg.runtime.logformat, level=logging.DEBUG, handlers=[console_logger])
    root_logger = logging.getLogger('')
    console_logger.terminator = ""

    logfile = os.path.join(cfg.general.result_path, cfg.general.error_log)
    file_logger = toolbox.SelectiveFileHandler(logfile, mode="w", DEBUG=cfg.runtime.DEBUG, ERROR=True)
    root_logger.addHandler(file_logger)

    # Write header for summary file.
    try:
        toolbox.save_csv(summary_path, [[ "Student", "LMS ID", "Score" ]])
    except Exception as e:
        logging.info("Warning: couldn't open summary file for writing: [%s]" % summary_path)

    # Run basic initialization routines.
    framework_context = prepare_environment_and_framework(cfg)

    # Close general log file and move on to student-specific logs.
    root_logger.removeHandler(file_logger)
    file_logger.close()

    # Prepare and run each submission.
    for submission in glob.glob(os.path.join(cfg.runtime.target_path, cfg.runtime.set)):
        submission_info = os.path.basename(submission).split("_", 1)
        student_name, lms_id = submission_info + ["NONE"] * (2 - len(submission_info))
        output_dir = os.path.join(cfg.general.result_path, os.path.basename(submission))

        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)

        logfile = os.path.join(output_dir, cfg.general.error_log)
        file_logger = toolbox.SelectiveFileHandler(logfile, mode="w", DEBUG=cfg.runtime.DEBUG, ERROR=True)
        root_logger.addHandler(file_logger)

        exec_class = pools.ThreadPool if cfg.runtime.threaded else pools.ProcessPool
        with exec_class() as executor:
            try:
                future = executor.apipe(prepare_and_test_submission, submission, framework_context, cfg)
                suite_results, exception_sets = future.get()

                # If there were exceptions in the tests, we should log them.
                for project, exception_list in exception_sets.items():
                   if len(exception_list) > 0:
                       log_header = "Exceptions for %s\n%s\n" % (project, "-"*(15 + len(project)))
                       logging.error(log_header + "\n".join(exception_list))
            except Exception as e:
                stack_trace = traceback.format_exc()
                logging.error("Error preparing / running %s - %s: %s\n%s" % (submission, type(e).__name__, e, stack_trace))
                root_logger.removeHandler(file_logger)
                file_logger.close()
#                executor.terminate()
#                executor.join()
                continue

#            executor.terminate()
#            executor.join()

        # Generate and save individual test score information to results file.
        grand_total = 0.0

        file_data = [["Scores for %s (LMS ID: %s)..." % (student_name, lms_id)]]

        if suite_results:
            for name, result, data_set in suite_results:
                file_data.extend([[]] + data_set + [["Set Total: %.3f" % result]])
                grand_total += result

        file_data.extend([[], ["Overall Score: %.2f" % grand_total]])
        toolbox.save_csv(os.path.join(output_dir, cfg.general.result_file), file_data)

        # Add data to summary file for this submission.
        try:
            toolbox.append_csv(summary_path, [[student_name, lms_id, grand_total]])
        except:
            # Fail silently; we should have already detected the error when creating the file.
            pass

        root_logger.removeHandler(file_logger)
        file_logger.close()
        time.sleep(2)

    # Shut everything down.
    shutdown_framework_and_environment(cfg, framework_context)

    logging.info("Framework shutdown\n")
    # Return to where we started at.
    os.chdir(starting_dir)


if __name__ == "__main__":
    main()

# Old alt version. Delete?
    # Initialize the framework / get any important info
#    with futures.ProcessPoolExecutor() as executor:
#        try:
#            future = executor.submit(prepare_and_init_framework, cfg)
#            framework_context = future.result()
            # This is a fatal error; if we can't initalize the framework, we should stop here.
#        except Exception as e:
#            sys.stderr.write("Error initializing framework - %s: %s. Exiting.\n" % (type(e).__name__, e))
#            exit()




import inspect
import logging
import os.path
from types import SimpleNamespace
from monkeydict import MonkeyDict


VERSION = '0.9.9.18'

# TODO: Make this prettier
def attr_has_value(base, attr_name):
    return True if hasattr(base, attr_name) and getattr(base, attr_name) else False


# Test Suite Configuration
class Config(MonkeyDict):
    _FRAMEWORK_BUILD_FLAGS = ( # Build flags for framework. All default to true
                     "prep_cmd",
                     "compile_cmd",
                     "post_cmd")

    _BUILD_PATHS = ( # Paths to base files and target source destination
                     "base",
                     "destination",
                     "resources",

                     # Build parameters: source folder (src), build destination (bin)
                     "subject_src",
                     "subject_bin",
                     "framework_src",
                     "framework_bin")

    _COMMAND_SETS = ( # Init / shutdown: command-line steps taken at startup and termination of test suite
                      "init_cmd",
                      "shutdown_cmd",

                      # Build commands: preparing & compiling (special path keys: $source_dir, $build_dir)
                      "prep_cmd",
                      "compile_cmd",
                      "post_cmd")
    # Command format: list[] is single command's elements (command and arguments); tuple() is list of commands.

    def __init__(self, runtime=None, test_sets=None, **keywords):
        super().__init__()
        self["runtime"] = runtime
        self["sets"] = test_sets

        # Log locations
        self.general = MonkeyDict({"result_path":  "Results",
                        "result_file":  "result.csv",
                        "error_log":    "error.log",
                        "summary_file": "summary.csv"})

        # Default build paths and command sets set to None, and build flags to True.
        self.build = MonkeyDict({entry: None for entry in (Config._BUILD_PATHS + Config._COMMAND_SETS)})
        self.build.framework_flags = MonkeyDict({entry: True for entry in Config._FRAMEWORK_BUILD_FLAGS})

        # Process specially recognized keywords
        if "threaded" in keywords:
            if self.runtime.threaded:
                logging.warning("WARNING: test suite threading config replaced by command line (threaded=True)")
                keywords.pop("threaded")
            else:
                self.runtime.threaded = keywords.pop("threaded")

        # Process keywords to make additional assignments.
        for key, value in keywords:
             self[key] = value


    # Normalize the configuration data (paths, etc.)
    def normalize(self):

        # Make configuration paths absolute
        self.general.result_path = os.path.abspath(self.general.result_path) if attr_has_value(self.general, "result_path") else None

        for entry in Config._BUILD_PATHS:
            self.build[entry] = os.path.abspath(self.build[entry]) if self.build[entry] else None

        # Verify / reformat all command sets in standard format (tuple of lists)
        for entry in Config._COMMAND_SETS:
            if self.build[entry]:
                self.build[entry] = self.build[entry] if isinstance(self.build[entry], tuple) else (self.build[entry],)

#        self.build.base = os.path.abspath(self.build.base) if attr_has_value(self.build, "base") else None
#        self.build.destination = os.path.abspath(self.build.destination) if attr_has_value(self.build, "destination") else None
#        self.build.resources = os.path.abspath(self.build.resources) if attr_has_value(self.build, "resources") else None
#        self.build.subject_src = os.path.abspath(self.build.subject_src) if attr_has_value(self.build, "subject_src") else None
#        self.build.subject_bin = os.path.abspath(self.build.subject_bin) if attr_has_value(self.build, "subject_bin") else None
#        self.build.framework_src = os.path.abspath(self.build.framework_src) if attr_has_value(self.build, "framework_src") else None
#        self.build.framework_bin = os.path.abspath(self.build.framework_bin) if attr_has_value(self.build, "framework_bin") else None


    ######################################################
    # Initialization and shutdown of components

    # Returns framework context
    def initialize_framework(self):
        return None


    def shutdown_framework(self, framework_context):
        return framework_context


    # Returns subject context
    def initialize_subject(self, submission, framework_context):
        return submission


    def shutdown_subject(self, subject_context):
        return subject_context


    # Returns test set context
    def initialize_test_set(self, test_set, subject_context, framework_context):
        return test_set


    def shutdown_test_set(self, test_set_context):
        return test_set_context



# Test Set (Subsection of Test Suite)
class TestSet:
    def __init__(self, name, id, num_tests, test_function, **keywords):
        self._name = name
        self._id = id

        # Assign the test count function. If it isn't valid, throw an exception.
        if isinstance(num_tests, int):
            self.get_num_tests = lambda *args, **kwargs: num_tests
        elif callable(num_tests):
            if len(inspect.getfullargspec(num_tests).args) == len(inspect.getfullargspec(TestSet.__num_tests_template).args):
               self.get_num_tests = num_tests
            else:
                raise Exception("Test count function has wrong number of paramters")
        else:
            raise Exception("Test count is invalid type (must be callable or integer)")

        # Assign the test run function. If it isn't valid, throw an exception.
        if callable(test_function):
            if len(inspect.getfullargspec(test_function).args) == len(inspect.getfullargspec(TestSet.__run_test_template).args):
                self.run_case_test = test_function
            else:
                raise Exception("Test run function has wrong number of paramters")
        else:
            raise Exception("Test run function is not callable")

        # Grab optional keyword argument values
        self._max_score = keywords.pop("max_score", 100.0)
        self._max_penalty = keywords.pop("max_penalty", 0.0)
        self.get_test_desc = keywords.pop("test_desc", TestSet.__get_test_description)

        # Initialize penalty lists
        self._case_penalties = []
        self._set_penalties = []
        #self.project_penalties = [ ("Time", 0.3) ]
        #self.test_case_penalties = [ ("Leaks", 0.3), ("Memory", 0.3) ]
        #self.max_penalty = 0.3
        #self._max_penalty = keywords.pop("max_penalty", 0.0)
        #self._set_penalties = keywords.pop("set_penalties", [])
        #self._case_penalties = keywords.pop("case_penalties", [])


    @property
    def name(self):
        return self._name


    @property
    def id(self):
        return self._id


    @property
    def num_tests(self):
        return self._num_tests


    @property
    def max_score(self):
        return self._max_score


    @property
    def case_penalties(self):
        return self._case_penalties


    @property
    def set_penalties(self):
        return self._set_penalties


    @property
    def max_penalty(self):
        return self._max_penalty


    @staticmethod
    def __num_tests_template(set_context, subject, framework, cfg):
        raise Exception("Template function should never be called!")


    @staticmethod
    def __run_test_template(test_num, set_context, subject, framework, cfg):
        raise Exception("Template function should never be called!")
        pass


    @staticmethod
    def __get_test_description(test_num, *arg_list, **keywords):
        return "Test #%d" % test_num


        #self.project_penalties = [ ("Time", 0.3) ]
        #self.test_case_penalties = [ ("Leaks", 0.3), ("Memory", 0.3) ]
    def add_case_penalty(self, name, fraction, function):
        if fraction and function:
            self._case_penalties.append((name, fraction, function))


    def add_set_penalty(self, name, fraction, function):
        if fraction and function:
            self._set_penalties.append((name, fraction, function))
